// ***************************************************************************
// AFLOW_COMPARE_STRUCTURE
// ***************************************************************************

#include<iostream>
#include "aflow_pflow.h"
#include "aflow.h"
#include "math.h"
#include<vector>
#include<string>
//#include <stdatomic.h> // Needed to communicate between threads
//[THREADS]#include<atomic>

using aurostd::isequal;
using aurostd::string2utype;
using aurostd::FileExist;

#ifndef __AFLOW_COMPARE_STRUCTURE_H_
#define __AFLOW_COMPARE_STRUCTURE_H_

#define JSON_MODE 0

// ===== StructurePrototype Class ===== //
class StructurePrototype{
  public:
    StructurePrototype();
    ~StructurePrototype();
    friend ostream& operator<<(ostream& oss, const StructurePrototype& StructurePrototype);
    const StructurePrototype& operator=(const StructurePrototype& b);
    StructurePrototype(const StructurePrototype& b);
    int iomode;
    string master_structure_name;
    xstructure master_structure;
    vector<string> elements;
    vector<int> stoichiometry;
    string pearson;
    uint space_group;
    vector<string> proto_structures_names;
    vector<xstructure> proto_structures;
    vector<double> misfits;
  private:
    void Free();
    void Copy(const StructurePrototype& b);
};

namespace compare{

  // ===== Main functions ===== //
  bool aflowCompareStructure(const int& num_proc, const xstructure& xstr1, const xstructure& xstr2, 
                           const bool& same_species, const bool& scale_volume, ostream& oss, double& final_misfit); //Main function //DX 12/2/17 added scale_volume
  //DX 12/2/17 : bool aflowCompareStructure(const int& num_proc, xstructure& xstr1, xstructure& xstr2, 
  //                             const bool& same_species, ostream& oss, double& final_misfit); //Main function
  bool aflowCompareStructure(const xstructure& xstr1, const xstructure& xstr2, const bool &same_species); //Overloaded, returns true (match), false (no match)
  bool aflowCompareStructure(const xstructure& xstr1, const xstructure& xstr2, const bool& same_species, const bool& scale_volume); 
  double aflowCompareStructureMisfit(const xstructure& xstr1, const xstructure& xstr2, const bool &same_species); //Overloaded, returns misfit value
//  string CompareStructures(aurostd::xoption& vpflow);
//  string CompareStructureDirectory(aurostd::xoption& vpflow);

  // ===== Compare Directory Functions ===== //
  vector<int> getStoichiometry(const xstructure& xstr, const bool& same_species);
  vector<string> getElements(xstructure& xstr);
  vector<int> gcdStoich(const deque<int>& numbers);
  void calculateSymmetry(xstructure& xstr, vector<string>& vpearsons, vector<uint>& vsgroups);
  void createStructurePrototypes(vector<StructurePrototype>& comparison_schemes, const vector<xstructure>& vxstrs, 
                                 const bool& same_species, const vector< vector<string> >& vvelements,
                                 vector< vector<int> >& vstoichs, vector<string>& vpearsons, vector<uint>& vsgroups, 
                                 const string& directory, const vector<string>& vfiles);
  int numberMismatches(const vector<StructurePrototype> comparison_schemes);
  void appendStructurePrototypes(vector<StructurePrototype>& comparison_schemes, 
                                 vector<StructurePrototype>& final_prototypes);
  void checkPrototypes(const int& num_proc, const bool& same_species, vector<StructurePrototype>& final_prototypes);
  void prepareTextOutput(ostream& ss_out, const bool& same_species, const vector<StructurePrototype>& final_prototypes);


  // ===== All other functions ===== //
  bool matchableSpecies(const xstructure& xstr1, const xstructure& xstr2, const bool& same_species);
  bool sameSpecies(const xstructure& x1, const xstructure& x2, const bool& display); 
  void rescaleStructure(xstructure& x1, xstructure& x2);
  void atomicNumberDensity(xstructure& xstr1, xstructure& xstr2);
  void fakeAtomsName(xstructure& x1);
  void printParameters(xstructure& xstr, ostream& oss);
  string leastFrequentAtom(const xstructure& xstr);
  vector<string> leastFrequentAtom2(const xstructure& xstr);
  bool checkTolerance(xvector<double> d1, xmatrix<double> Q2);
  bool checkABCTolerance(xvector<double> d1, xvector<double> d2);
  bool checkAngleTolerance(xvector<double> d1, xvector<double> d2);
  void findMatch(const xstructure& xstr1, const xstructure& PROTO,vector<uint>& im1, vector<uint>& im2);
  void clusterize(const xstructure& xstr1, const vector<uint>& im1, vector<string>& TYPE1,
                  vector<uint>& QTA1, vector<vector<uint> >& I1);
  bool sameAtomType(const xstructure& xstr1, const xstructure& xstr2, const vector<uint>& im1,
                    const vector<uint>& im2, const int& type_match);
  bool cleanMatch(const vector<uint>& im1);
  void cellDiagonal(const xstructure& xstr, vector<double>& diag_sum, vector<double>& diag_diff, const double& scale);
  double latticeDeviation(const vector<double>& diag_sum1,const vector<double>& diag_sum2, 
                          const vector<double>& diag_diff1,const vector<double>& diag_diff2); 
  double shortestDistance(const xstructure& xstr, const uint& k);
  void coordinateDeviation(const xstructure& xstr1, const xstructure& xstr2,
                      const vector<uint>& indexMatch1, const vector<uint>& indexMatch2,
                      double& cd, double& fail_figure);
  double computeMisfit(const double& dev, const double& dis, const double& fail);
  void printMatch(const vector<uint>& indexMatch1, const vector<uint>& indexMatch2,
                  const xstructure& PROTO, const xstructure& xstr1, ostream& oss);
  xvector<double> bringCoordinateInCell(xvector<double>& coord);
  bool quadrupletPeriodic(const xmatrix<double>& quad, const xstructure& lfa_supercell, const int& i, 
                       const int& j, const int& k, const int& w);
  void threadGeneration(const int& num_proc,const xmatrix<double>& q1, xstructure& xstr2, 
                        vector<xstructure> &vprotos, xstructure &xstr1, const int& type_match, 
                        double& minMis, ostream& oss);
//[THREADS]  void quadrupletSearch(std::atomic_bool &misfit_in_threshold_found, const string& lfa, const int& min_index_i, 
//[THREADS]                        const int& max_index_i,const int& min_index_j, const int& max_index_j,const int& min_index_k, 
//[THREADS]                        const int& max_index_k, const int& min_index_w,const int& max_index_w, const xmatrix<double>& q1, 
//[THREADS]                        const xstructure& xstr, const xstructure& xstr_LFA_supercell, vector<xstructure>& vprotos, 
//[THREADS]                        xstructure& xstr1, const xstructure& xstr2, const int& type_match, double& possible_minMis, ostream& oss);
  void quadrupletSearch(bool &misfit_in_threshold_found, const string& lfa, const uint& min_index_i,
                        const uint& max_index_i,const uint& min_index_j, const uint& max_index_j,const uint& min_index_k,
                        const uint& max_index_k, const uint& min_index_w,const uint& max_index_w, const xmatrix<double>& q1,
                        const xstructure& xstr, const xstructure& xstr_LFA_supercell, vector<xstructure>& vprotos,
                        xstructure& xstr1, const xstructure& xstr2, const uint& type_match, double& possible_minMis);

} // end namespace

#endif 

// ***************************************************************************
// AFLOW_COMPARE_STRUCTURE
// ***************************************************************************

